#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Constants used in this module.

"""

# Database connexion timeout (in sec)
TIMEOUT = 120

# Default maximum threads number
MAX_THREADS_DEFAULT = 1

# Errors that return exit code = 1
EXIT_ERRORS = ['003', '005']
